'''
Computes the 3S fixed schedule of solvers within a time-out of 
C seconds, using precomputed coefficient matrices and headers.
'''

import sys
import os
import csv
from gurobipy import *
from time import clock
import json

def get_partial_score(instance, solver, time):
  item = infos[instance][solver]
  vals = item['vals']
  values = [int(v) for (v, t) in vals.items() if t <= time]
  if not values:
    return 0
  if item['obj'] == 'min':
    val = min(values)
  else:
    val = max(values)
  return compute_score(
    item['info'], 
    item['min_val'], 
    item['max_val'], 
    item['obj'], 
    val,
    time,
    item['time']
  )

def compute_score(info, min_val, max_val, obj, val, time, tot_time):
  if info == "unknown":
    return 0
  if info != "sat" and time >= tot_time:
    return 1
  if min_val == max_val:
    return 0.75
  s = ((val - min_val) * 0.5) / (max_val - min_val)
  if obj == "min":
    return 0.75 - s
  if obj == "max":
    return s + 0.25
  #print "Error!"
  system.exit(1)

with open('/import/atp-linhome/home/1/ramadini/bologna/cop/3S/infos', 'r') as infile:
  infos = json.load(infile)

def main(args):    
        C = 180.0
        in_path = '/import/atp-linhome/home/1/ramadini/bologna/cop/3S/'
        #out_file = csv.writer(open('./schedule', "ab"), delimiter='|', skipinitialspace = True)
        num_solvers = 12
  
        n_solver = args[0]
        rep = args[1]
        fold = args[2]
	start = clock()
	name = 'rep' + str(rep) + '_fold' + str(fold)
	#print 'Building model ' + name + '...'
	model = Model(name)
        model.setParam('OutputFlag', False)
	path = in_path + str(n_solver) + '_solvers/'
	# coeff is a coefficient matrix of size (N + 1) * M.
	coeff  = csv.reader(open(path + 'coeff_train_'  + name, "rb"), delimiter=" ", skipinitialspace = True)
	# header is a row of length M which contains the names of the variables.
	header = csv.reader(open(path + 'header_train_' + name, "rb"), delimiter=" ", skipinitialspace = True)
	
	#print 'Adding binary variables to the model...'
	coeff_list = list(coeff)
	N = len(coeff_list) - 1
	header_list = list(header)[0]
	# The last element of the header (and the last column of the coefficient matrix) is actually a blank space.
	M = len(header_list) - 1
	x = [Var] * M
	for i in range(0, N):
	  # header_list[i] (i.e. the name of the instance) is too long.
	  x[i] = model.addVar(vtype = GRB.BINARY, name = 'y_' + str(i))
	for j in range(N, M):
	  x[j] = model.addVar(vtype = GRB.BINARY, name = header_list[j])
	model.update()
	
	#print 'Adding contraints to the model...'
	for i in range(0, N):
	  row_vars = []
	  for j in range(N, M):
	    if int(coeff_list[i][j]) == 1:
	      row_vars.append(x[j])
	  model.addConstr(x[i] + quicksum(row_vars) >= 1)
	  
	exprs  = []
	for k in range(N, M):
	  t = coeff_list[N][k]
	  exprs.append(LinExpr(float(t), x[k]))
	model.addConstr(quicksum(exprs) <= C)
	
	#print 'Adding objective function to the model...'
        exprs = []
	max_f = 0
	for i in range(0, N):
	  inst = header_list[i]
	  for j in range(N, M):
	    solver_time = header_list[j]
	    k = solver_time.index('_')
	    solver = solver_time[0 : k]
	    time = float(solver_time[k + 1 : len(solver_time)])
	    score = get_partial_score(inst, solver, time)
	    if score == 0 or score == 1:
	      f = time
	    else:
	      f = C + (C / score) + (time / C)
	    if f > max_f:
	      max_f = f	    
	    exprs.append(LinExpr(f / N, x[j]))
	max_f += 1
        for i in range(0, N):
	  exprs.append(LinExpr(max_f, x[i]))
	model.setObjective(quicksum(exprs), GRB.MINIMIZE)
	
	#print 'Solving the model...'
	model.optimize()
	elapsed = (clock() - start)
	#print 'Objective value:', model.objVal
	#print 'Schedule results:'
	sol = model.getVars()
	#print 'Solution length: ' + str(len(sol))
	k = 0
	for i in range(0, N):
	  if sol[i].x == 1:
	    k += 1
	    #print "Solver schedule can not solve instance: " + header_list[i]
	#print "Total no. of instances that solver can't solve: " + str(k) + "/" + str(N)
	
	k = 0
	for j in range(N, M):
	  if round(sol[j].x) == 1:
	    solver_time = sol[j].varName
	    k = solver_time.index('_')
	    solver = solver_time[0 : k]
	    time = solver_time[k + 1 : len(solver_time)]
	    #print "Solver schedule will run: " + solver_time
	    #line = [n_solver, rep, fold, solver, time, elapsed]
            #out_file.writerow(line)
            print n_solver + '|' + rep + '|' + fold + '|' + solver + '|' + str(time) + '|' + str(elapsed)
	  
if __name__ == '__main__':
    main(sys.argv[1:])
